# Author: Scott Woods <scott.18.ansar@gmail.com.com>
# MIT License
#
# Copyright (c) 2017-2023 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
__docformat__ = 'restructuredtext'

from .socketry import *

__all__ = [
	'ANSAR_LOCAL_HOST',
	'ANSAR_LAN_HOST',
	'ANSAR_DIRECTORY_PORT',
	'ANSAR_RESERVED_PORT',
	'ANSAR_DEDICATED_PORT',
	'ANSAR_SHARED_PORT',
	'ANSAR_LOCAL_RESERVED',
	'ANSAR_LOCAL_DEDICATED',
	'ANSAR_LOCAL_SHARED',
	'ANSAR_LAN_RESERVED',
	'ANSAR_LAN_DEDICATED',
	'ANSAR_LAN_SHARED',
]

ANSAR_LOCAL_HOST = '127.0.0.1'
ANSAR_LAN_HOST = '192.168.1.176'

ANSAR_DIRECTORY_PORT = 32176
ANSAR_RESERVED_PORT = ANSAR_DIRECTORY_PORT
ANSAR_DEDICATED_PORT = ANSAR_DIRECTORY_PORT + 1
ANSAR_SHARED_PORT = ANSAR_DIRECTORY_PORT + 2

ANSAR_LOCAL_RESERVED = HostPort(ANSAR_LOCAL_HOST, ANSAR_RESERVED_PORT)
ANSAR_LOCAL_DEDICATED = HostPort(ANSAR_LOCAL_HOST, ANSAR_DEDICATED_PORT)
ANSAR_LOCAL_SHARED = HostPort(ANSAR_LOCAL_HOST, ANSAR_SHARED_PORT)
ANSAR_LAN_RESERVED = HostPort(ANSAR_LAN_HOST, ANSAR_RESERVED_PORT)
ANSAR_LAN_DEDICATED = HostPort(ANSAR_LAN_HOST, ANSAR_DEDICATED_PORT)
ANSAR_LAN_SHARED = HostPort(ANSAR_LAN_HOST, ANSAR_SHARED_PORT)



# Author: Scott Woods <scott.18.ansar@gmail.com.com>
# MIT License
#
# Copyright (c) 2017-2023 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Tools and runtime for asynchronous programming.

Repo: git@github.com:mr-ansar/ansar-connect.git
Branch: main
Commit: c89b59a00edd5d39223f7f796ac2ef0b033e2d12
Version: 0.1.255 (2024-05-22@13:15:16+NZST)
"""

from ansar.create import *

#bind = bind_any
#create = create_object

from .socketry import HostPort, LocalPort
from .socketry import ScopeOfIP, local_private_public
from .socketry import Blob, CreateFrame
from .socketry import Listening, NotListening, Accepted, NotAccepted, StopListening
from .socketry import Connected, NotConnected
from .socketry import Close, Closed, Abandoned
from .transporting import listen, connect, stop_listen

from .plumbing import RETRY_LOCAL, RETRY_PRIVATE, RETRY_PUBLIC
from .plumbing import ip_retry

from .directory_if import ScopeOfService
from .directory_if import Published, NotPublished, Subscribed
from .directory_if import Available, NotAvailable, Delivered, NotDelivered
from .directory_if import Clear, Cleared, Dropped
from .directory_if import NetworkEnquiry, NetworkConnect, DirectoryScope, DirectoryAncestry

from .directory import ServiceDirectory
from .directory import RouteByRelay, InboundByRelay, OpenLoop, CloseLoop
from .directory import publish, subscribe
from .directory import clear, retract
from .directory import key_service

from .networking_if import UseAddress, NoAddress, GlareTimer
from .networking import ConnectToAddress, ListenAtAddress
from .networking import SubscribeToListing, PublishAListing, SubscribeToSearch
from .grouping import GroupTable, GroupUpdate, AddressGroup, GroupTimer, GroupObject

from .node import NodeSettings, node_settings
from .node import node_passing, sub_node_passing
from .node import create_node, NodeProperties

from .moving import overlay, mirror

from .wan import CONTACT_TYPE, CONTACT_DEVICE
from .wan import EmailAddress, PhoneNumber



# Author: Scott Woods <scott.18.ansar@gmail.com.com>
# MIT License
#
# Copyright (c) 2017-2024 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
'''A network fixture that creates directories for connected clients.

Rather than a unique network address (IP + port) per directory, this
process uses a single address while managing multiple directories. Multiple
deployments of different products or different instances of the same
product can be supported.
'''
import ansar.connect as ar
from ansar.connect.standard import *
from ansar.connect.wan import *
from ansar.connect.product import *
from ansar.connect.directory import find_overlap

__all__ = [
	'main',
]

# All the states for all the machines in
# this module.
class INITIAL: pass
class STARTING: pass
class READY: pass
class CLEARING: pass

CONNECT_ABOVE_TABLE = ar.MapOf(ar.Unicode(), ar.Any())

# The process object.
#
class ProductDirectory(ar.Threaded, ar.StateMachine):
	def __init__(self, settings):
		ar.Threaded.__init__(self)
		ar.StateMachine.__init__(self, INITIAL)
		self.settings = settings
		self.public_access = None		# This fixture listens here.
		self.accepted = {}				# Connected clients.
		self.directory = {}				# Directory per product+instance.
		self.connect_above = {}			# Upward link per product+instance.
		self.child_of = {}			# Sub-directories to directory.
		self.model = None				# File storage of connect_above.
		self.store = None				# Storage methods.
		self.recover = None

def ProductDirectory_INITIAL_Start(self, message):
	model = ar.object_model_folder()
	if model is None:
		self.complete(ar.Failed(model_load=(None, 'no model storage')))

	self.model = model.file('connect-above', CONNECT_ABOVE_TABLE, create_default=True)

	# Custom I/O routines.
	def store(connect_above=None):
		connect_above = connect_above or self.connect_above
		self.model.store(connect_above)

	def recover():
		self.connect_above, _ = self.model.recover()
		return self.connect_above
	
	self.store = store
	self.recover = recover

	# Initial load.
	self.recover()

	settings = self.settings
	encrypted = settings.encrypted
	if settings.custom_port:
		d = settings.public_access.port - ANSAR_DIRECTORY_PORT
		settings.public_access.port = settings.custom_port + d
	ar.listen(self, settings.public_access, encrypted=encrypted)
	return STARTING

# Verify access
def ProductDirectory_STARTING_Listening(self, message):
	self.public_access = message
	return READY

def ProductDirectory_STARTING_NotListening(self, message):
	# Something wrong. Bail.
	self.complete(ar.Faulted('No public access for product/instance', message.error_text))

def ProductDirectory_STARTING_Stop(self, message):
	# External intervention.
	self.complete(ar.Aborted())

# Connections from IP clients.
def ProductDirectory_READY_Accepted(self, message):
	# Need to remember connection details for full
	# operation of directories. Accepted messages
	# will be forwarded later to appropriate
	# directories.
	r = self.return_address[-1]
	self.accepted[r] = message

	self.trace(f'Accepted child [{r}] ({message.accepted_ipp})')
	return READY

def ProductDirectory_READY_Abandoned(self, message):
	# Need to generate abandon messages for the
	# relevant directories.
	r = self.return_address[-1]
	p = self.accepted.pop(r, None)
	if p is None:
		self.warning(f'Abandoned by unknown child [{r}]')
		return READY

	hap = str(p.accepted_ipp)
	d = self.child_of.pop(r, None)
	if d is None:
		self.trace(f'Abandoned by child [{r}] ({hap}), not registered as sub-directory')
		return READY

	self.forward(message, d, self.return_address)
	self.trace(f'Abandoned by child [{r}] ({hap})')

	return READY

def ProductDirectory_READY_Closed(self, message):
	# Need to generate abandon messages for the
	# relevant directories.
	r = self.return_address[-1]
	p = self.accepted.pop(r, None)
	if p is None:
		self.warning(f'Close of unknown child [{r}]')
		return READY

	hap = str(p.accepted_ipp)
	d = self.child_of.pop(r, None)
	if d is None:
		self.trace(f'Close of child [{r}] ({hap})" (not registered as sub-directory')
		return READY

	self.forward(message, d, self.return_address)
	self.trace(f'Close of child [{r}] ({hap})')

	return READY

def ProductDirectory_READY_ProductLookup(self, message):
	r = self.return_address[-1]

	settings = self.settings

	product_name = message.product_name or 'Ansar Networking'
	product_instance = message.product_instance or InstanceOfProduct.TESTING
	k = f'{product_name}:{product_instance}'

	d = self.directory.get(k)
	if d is None:
		self.trace(f'Unknown directory key "{k}" for child [{r}]')
		c = self.connect_above.get(k)
		if c is None:
			self.trace(f'Unknown upward key "{k}" for child [{r}]')
			# Default upward access is disabled but a hint that it
			# should adopt the same nature.
			c = ProductAccess(product_name=product_name, product_instance=product_instance)
			self.connect_above[k] = c
			self.store()

		d = self.create(ar.ServiceDirectory,
			scope=settings.directory_scope,			# All at same scope.
			connect_above=c,
			accept_below=ar.HostPort())				# Disabled.

		self.assign(d, k)
		self.directory[k] = d

	self.child_of[r] = d
	accepted = self.accepted.get(r)
	if accepted:
		self.forward(accepted, d, self.return_address)
	else:
		self.trace(f'Not an accepted child [{r}]')
	self.reply(YourProduct(d))
	return READY

def ProductDirectory_READY_HostPort(self, message):
	# Consume the redundant message. Listening by
	# all directories is disabled, so host-port
	# is empty.
	return READY

def ProductDirectory_READY_Anything(self, message):
	connect_above = message.thing

	if not isinstance(connect_above, (ProductAccess, WideAreaAccess)):
		t = ar.tof(connect_above)
		self.warning(f'Attempt to connect up to "{t}" from a product directory')
		self.reply(ar.Ack())
		return READY

	k = self.progress()
	if k is None:
		self.warning(f'Attempt to connect up from an unknown directory')
		self.reply(ar.Ack())
		return READY

	self.connect_above[k] = connect_above
	self.store()

	self.reply(ar.Ack())
	return READY

def ProductDirectory_READY_Completed(self, message):
	k = self.debrief()
	d = self.directory.pop(k, None)
	if d is None:
		self.warning(f'Termination of unknown object')
		return READY
	self.warning(f'Termination of directory "{k}", tearing down')
	
	# A directory has died. Force any sub-directories
	# to re-connect and thereby create a new
	# instance.
	c = []
	for s, v in self.child_of.items():
		if v == d:
			self.send(ar.Close(), (s,))
			c.append(s)

	for t in c:
		self.child_of.pop(t, None)
	return READY

def ProductDirectory_READY_Stop(self, message):
	if self.working():
		self.abort()
		return CLEARING
	self.complete(ar.Ack())

def ProductDirectory_CLEARING_Completed(self, message):
	self.debrief()
	if self.working():
		return CLEARING
	self.complete(ar.Ack())

PRODUCT_DIRECTORY_DISPATCH = {
    INITIAL: (
        (ar.Start,), ()
    ),
    STARTING: (
        (ar.Listening, ar.NotListening, ar.Stop), ()
    ),
    READY: (
        (ar.Accepted, ar.Abandoned, ar.Closed,
		ProductLookup, ar.HostPort,
		ar.Anything,
		ar.Completed,
		ar.Stop), ()
    ),
    CLEARING: (
        (ar.Completed), ()
    ),
}

ar.bind(ProductDirectory, PRODUCT_DIRECTORY_DISPATCH)


# Allow configuration of network details.
#
class Settings(object):
	def __init__(self, directory_scope=None, public_access=None, encrypted=False, custom_port=None):
		self.directory_scope = directory_scope
		self.public_access = public_access or ar.HostPort()
		self.encrypted = encrypted
		self.custom_port = custom_port

SETTINGS_SCHEMA = {
	'directory_scope': ar.ScopeOfService,
	'public_access': ar.UserDefined(ar.HostPort),
	'encrypted': ar.Boolean(),
	'custom_port': ar.Integer8(),
}

ar.bind(Settings, object_schema=SETTINGS_SCHEMA)

#
#
factory_settings = Settings(directory_scope=ar.ScopeOfService.HOST,
	public_access=ANSAR_LOCAL_SHARED,
	custom_port=0)

#
#
def main():
    ar.create_object(ProductDirectory, factory_settings=factory_settings)

# The standard entry point. Needed for IDEs
# and debugger sessions.
if __name__ == '__main__':
    main()



# Author: Scott Woods <scott.18.ansar@gmail.com.com>
# MIT License
#
# Copyright (c) 2017-2023 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
'''Interface for the cloud. Needed for ansar CLI.

Defined here so that dependency is from cloud to ansar-connect, rather
than linking ansar-connect to the cloud. Cloud codebase not dragged into
this library.
'''

import ansar.connect as ar
from ansar.connect.product import *

__all__ = [
	'FOH_PORT',
	'AccountSignup',
	'AccountLogin',
	'AccountRead',
	'AccountUpdate',
	'AccountDelete',
	'AccountAddLogin',
	'AccountDeleteLogin',
	'AccountAddDirectory',
	'AccountDeleteDirectory',
	'LoginRead',
	'DirectoryRead',
	'DirectoryUpdate',
	'DirectoryExport',
	'AccountOpened',
	'DirectoryExported',
	'DirectoryPage',
	'AccountPage',
	'LoginPage',
	'LoginDeleted',
	'AccountDeleted',
	'DirectoryDeleted',
]

FOH_PORT=5022

# Forms to present to the cloud foh.
# Open an existing session with an account/developer
class AccountSignup(object):
	def __init__(self, login_email=None, password=None,
			family_name=None, given_name=None, nick_name=None, honorific=None,
			organization_name=None, organization_location=None):
		self.login_email = login_email
		self.password = password
		self.family_name = family_name
		self.given_name = given_name
		self.nick_name = nick_name
		self.honorific = honorific
		self.organization_name = organization_name
		self.organization_location = organization_location

class AccountLogin(object):
	def __init__(self, login_email=None, password=None):
		self.login_email = login_email
		self.password = password

class AccountRead(object):
	def __init__(self, login_token=None):
		self.login_token = login_token

class AccountUpdate(object):
	def __init__(self, login_token=None, organization_name=None, organization_location=None):
		self.login_token = login_token
		self.organization_name = organization_name
		self.organization_location = organization_location

class AccountDelete(object):
	def __init__(self, login_token=None):
		self.login_token = login_token

class AccountAddLogin(object):
	def __init__(self, login_token=None, login_email=None, password=None,
			family_name=None, given_name=None, nick_name=None, honorific=None):
		self.login_token = login_token
		self.login_email = login_email
		self.password = password
		self.family_name = family_name
		self.given_name = given_name
		self.nick_name = nick_name
		self.honorific = honorific

class AccountDeleteLogin(object):
	def __init__(self, login_token=None, login_id=None):
		self.login_token = login_token
		self.login_id = login_id

class AccountAddDirectory(object):
	def __init__(self, login_token=None,
			product_name=None, product_instance=None):
		self.login_token = login_token
		self.product_name = product_name
		self.product_instance = product_instance

class AccountDeleteDirectory(object):
	def __init__(self, login_token=None, directory_id=None):
		self.login_token = login_token
		self.directory_id = directory_id

class LoginRead(object):
	def __init__(self, login_token=None, login_id=None):
		self.login_token = login_token
		self.login_id = login_id

class DirectoryRead(object):
	def __init__(self, login_token=None, directory_id=None):
		self.login_token = login_token
		self.directory_id = directory_id

class DirectoryUpdate(object):
	def __init__(self, login_token=None, directory_id=None):
		self.login_token = login_token
		self.directory_id = directory_id

class DirectoryExport(object):
	def __init__(self, login_token=None, directory_id=None, access_name=None):
		self.login_token = login_token
		self.directory_id = directory_id
		self.access_name = access_name

SHARED_SCHEMA = {
	"login_token": ar.UUID(),
	"login_email": ar.Unicode(),
	"password": ar.Unicode(),
	"family_name": ar.Unicode(),
	"given_name": ar.Unicode(),
	"nick_name": ar.Unicode(),
	"honorific": ar.Unicode(),
	"organization_name": ar.Unicode(),
	"organization_location": ar.Unicode(),
	"account_id": ar.UUID(),
	"product_name": ar.Unicode(),
	"product_instance": InstanceOfProduct,
	"directory_id": ar.UUID(),
	"exported_access": ar.VectorOf(ar.Unicode()),
	"number_of_logins": ar.Integer8(),
	"number_of_directories": ar.Integer8(),
	"number_of_relays": ar.Integer8(),
	"number_of_tokens": ar.Integer8(),
	"connected_routes": ar.Integer8(),
	"messages_per_second": ar.Integer8(),
	"bytes_per_second": ar.Integer8(),
	"exported_name": ar.VectorOf(ar.Unicode()),
	"directory_id": ar.UUID(),
	"account_id": ar.UUID(),
	"technical_contact": ar.VectorOf(ar.Any()),
	"financial_contact": ar.VectorOf(ar.Any()),
	"administrative_contact": ar.VectorOf(ar.Any()),
	"login_id": ar.UUID(),
	"assigned_directory": ar.SetOf(ar.UUID()),
	"access_token": ar.UUID(),
	"access_name": ar.Unicode(),
	"created": ar.WorldTime(),
}

ar.bind(AccountSignup, object_schema=SHARED_SCHEMA)
ar.bind(AccountLogin, object_schema=SHARED_SCHEMA)
ar.bind(AccountRead, object_schema=SHARED_SCHEMA)
ar.bind(AccountUpdate, object_schema=SHARED_SCHEMA)
ar.bind(AccountDelete, object_schema=SHARED_SCHEMA)
ar.bind(AccountAddLogin, object_schema=SHARED_SCHEMA)
ar.bind(AccountDeleteLogin, object_schema=SHARED_SCHEMA)
ar.bind(AccountAddDirectory, object_schema=SHARED_SCHEMA)
ar.bind(AccountDeleteDirectory, object_schema=SHARED_SCHEMA)
ar.bind(LoginRead, object_schema=SHARED_SCHEMA)
ar.bind(DirectoryRead, object_schema=SHARED_SCHEMA)
ar.bind(DirectoryUpdate, object_schema=SHARED_SCHEMA)
ar.bind(DirectoryExport, object_schema=SHARED_SCHEMA)

# Forms and responses to present to cloud clients.
class AccountOpened(object):
	def __init__(self, login_token=None):
		self.login_token = login_token

# Read/update response covered by AccountPage.

class LoginDeleted(object): pass
class AccountDeleted(object): pass
class DirectoryDeleted(object): pass

class DirectoryExported(object):
	def __init__(self, directory_id=None, access_token=None, account_id=None,
			product_name=None, product_instance=None):
		self.directory_id = directory_id
		self.access_token = access_token
		self.account_id = account_id
		self.product_name = product_name
		self.product_instance = product_instance

class DirectoryPage(object):
	def __init__(self, directory_id=None, product_name=None, product_instance=None,
			number_of_tokens=None, connected_routes=None, messages_per_second=None, bytes_per_second=None,
			exported_name=None, created=None):
		self.directory_id = directory_id			# Unique key.
		self.product_name = product_name
		self.product_instance = product_instance
		self.number_of_tokens = number_of_tokens
		self.connected_routes = connected_routes
		self.messages_per_second = messages_per_second
		self.bytes_per_second = bytes_per_second
		self.exported_name = exported_name or ar.default_vector()
		self.created = created

	def __str__(self):
		s = InstanceOfProduct.to_name(self.product_instance)
		return f'{self.product_name}/{s}'

class AccountPage(object):
	def __init__(self, account_id=None,
			organization_name=None, organization_location=None, technical_contact=None, financial_contact=None, administrative_contact=None,
			number_of_logins=None, number_of_directories=None, number_of_relays=None, created=None,
			login_page=None, directory_page=None):
		self.account_id = account_id				# Unique key.
		self.organization_name = organization_name
		self.organization_location = organization_location
		self.technical_contact = technical_contact or ar.default_vector()
		self.financial_contact = financial_contact or ar.default_vector()
		self.administrative_contact = administrative_contact or ar.default_vector()
		self.number_of_logins = number_of_logins
		self.number_of_directories = number_of_directories
		self.number_of_relays = number_of_relays
		self.created = created
		# Payment
		# Invoices
		self.login_page = login_page or ar.default_vector()
		self.directory_page = directory_page or ar.default_vector()

	def __str__(self):
		return f'{self.organization_name}'

class LoginPage(object):
	def __init__(self, login_id=None, login_email=None,
			account_id=None, assigned_directory=None,
			family_name=None, given_name=None, nick_name=None, honorific=None,
			created=None):
		self.login_id = login_id			# Unique key.
		self.login_email = login_email
		self.account_id = account_id		# Belong to this account.
		self.assigned_directory = assigned_directory or ar.default_set()	# Assigned use of.
		self.family_name = family_name
		self.given_name = given_name
		self.nick_name = nick_name
		self.honorific = honorific
		self.created = created

	def __str__(self):
		if self.nick_name:
			return self.nick_name

		if self.honorific:
			h = f'{self.honorific} '
		else:
			h = ''

		if self.family_name:
			if self.given_name:
				return f'{h}{self.given_name} {self.family_name}'
			return f'{h}{self.family_name}'
		elif self.given_name:
				return self.given_name

		return self.login_email

ar.bind(AccountOpened, object_schema=SHARED_SCHEMA)
ar.bind(DirectoryExported, object_schema=SHARED_SCHEMA)
ar.bind(LoginPage, object_schema=SHARED_SCHEMA)
ar.bind(DirectoryPage, object_schema=SHARED_SCHEMA)

DIRECTORY_SCHEMA = {
	"login_page": ar.VectorOf(ar.UserDefined(LoginPage)),
	"directory_page": ar.VectorOf(ar.UserDefined(DirectoryPage)),
}
DIRECTORY_SCHEMA.update(SHARED_SCHEMA)

ar.bind(AccountPage, object_schema=DIRECTORY_SCHEMA)
ar.bind(LoginDeleted, object_schema=SHARED_SCHEMA)
ar.bind(AccountDeleted, object_schema=SHARED_SCHEMA)
ar.bind(DirectoryDeleted, object_schema=SHARED_SCHEMA)
ar.bind(DirectoryDeleted, object_schema=SHARED_SCHEMA)



# Standard PyPi packaging.
# Build materials and push to pypi.org.
# Author: Scott Woods <scott.18.ansar@gmail.com.com>
import sys
import setuptools
import re

#
#
VERSION_PATTERN = re.compile(r'([0-9]+)\.([0-9]+)\.([0-9]+)')

#
#
with open("PACKAGE", "r", encoding="utf-8") as f:
    p = f.read()
PACKAGE = p[:-1]

#
#
with open("DESCRIPTION", "r", encoding="utf-8") as f:
    d = f.read()
DESCRIPTION = d[:-1]

#
#
with open("VERSION", "r", encoding="utf-8") as f:
    line = [t for t in f]
VERSION = line[-1][:-1]

if not VERSION_PATTERN.match(VERSION):
    print('Version "%s" does not meet semantic requirements' % (VERSION,))
    sys.exit(1)

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

with open("DOC_LATEST_LINK", "r", encoding="utf-8") as f:
    d = f.read()
DOC_LINK = d[:-1]

REQUIRES = [
    "cffi>=1.16.0",
    "PyNaCl>=1.5.0",
    "ansar-create>=0.1.89",
]

setuptools.setup(
    name=PACKAGE,
    version=VERSION,
    author="Scott Woods",
    author_email="scott.18.ansar@gmail.com.com",
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    #project_urls={
    #    "Documentation": DOC_LINK,
    #},
    #classifiers=[
    #    "Development Status :: 4 - Beta",
    #    "Intended Audience :: Developers",
    #    "Programming Language :: Python :: 3",
    #    "License :: OSI Approved :: MIT License",
    #    "Operating System :: OS Independent",
    #    "Topic :: Software Development :: Libraries",
    #],
    # Where multiple packages might be found, esp if using standard
    # layout for "find_packages".
    package_dir={
        "": "src",
    },
    # First folder under "where" defines the name of the
    # namespace. Folders under that (with __init__.py files)
    # define import packages under that namespace.
    packages=setuptools.find_namespace_packages(
        where="src",
    ),
    entry_points = {
        'console_scripts': [
            'ansar=ansar.command.ansar_command:main',
            'ansar-group=ansar.command.ansar_group:main',
            'ansar-directory=ansar.command.ansar_directory:main',
            'shared-directory=ansar.command.shared_directory:main',
        ],
    },
)



# Author: Scott Woods <scott.18.ansar@gmail.com.com>
# MIT License
#
# Copyright (c) 2017-2023 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
__docformat__ = 'restructuredtext'

import ansar.create as ar
from .socketry import *
from .transporting import *
from .networking_if import *
from .product import *
from .wan import *
from .plumbing import ip_retry

__all__ = [
	'ConnectToDirectory',
]

#
#
class INITIAL: pass
class DISABLED: pass
class PENDING: pass
class CONNECTED: pass
class PRODUCING: pass
class GLANCING: pass
class CLOUDY: pass
class INTERRUPTION: pass
class LOOKING: pass
class RECONNECTING: pass
class GLARING: pass
class CONNECTING: pass
class REDIRECTING: pass
class ASSIGNING: pass
class CLOSING: pass

class ConnectToDirectory(ar.Point, ar.StateMachine):
	def __init__(self, connect_above, session=None, group_address=None):
		ar.Point.__init__(self)
		ar.StateMachine.__init__(self, INITIAL)
		self.connect_above = connect_above
		self.session = session
		self.group_address = group_address

		self.started = None
		self.attempts = 0

		self.connected = None
		self.remote = None
		self.redirect = None

		self.closing = None
		self.intervals = None
		self.retry = None

	def connect(self):
		c = self.connect_above
		if isinstance(c, HostPort):
			ipp = c
			encrypted = False
			next = PENDING
		elif isinstance(c, ProductAccess):
			ipp = c.access_ipp
			encrypted = c.encrypted
			next = PRODUCING
		elif isinstance(c, WideAreaAccess):
			ipp = c.access_ipp
			encrypted = c.encrypted
			next = CLOUDY
		else:
			return DISABLED

		if ipp.host is None:
			return DISABLED

		connect(self, ipp, session=self.session, encrypted=encrypted)
		self.start(ar.T1, seconds=8.0)
		return next

	def reschedule(self):
		if self.intervals is None:
			c = self.connect_above
			if isinstance(c, HostPort):
				ipp = c
			elif isinstance(c, ProductAccess):
				ipp = c.access_ipp
			elif isinstance(c, WideAreaAccess):
				ipp = c.access_ipp
			else:
				ipp = None

			if ipp is None or ipp.host is None:
				s = ScopeOfIP.PUBLIC
			else:
				s = local_private_public(ipp.host)
			r = ip_retry(s)
			self.intervals = r
		
		if self.retry is None:
			self.retry = ar.smart_intervals(self.intervals)

		try:
			p = next(self.retry)
		except StopIteration:
			self.retry = None
			return False
	
		self.start(GlareTimer, p)
		return True

# INITIAL
# Launch this object.
def ConnectToDirectory_INITIAL_Start(self, message):
	self.group_address = self.group_address or self.parent_address

	# Start from nothing.
	self.started = ar.world_now()

	self.attempts = 1
	return self.connect()

# DISABLED
def ConnectToDirectory_DISABLED_Anything(self, message):
	self.connect_above = message.thing
	self.attempts = 1
	return self.connect()

def ConnectToDirectory_DISABLED_Stop(self, message):
	self.complete(ar.Aborted())

# PENDING (HostPort)
# Waiting for results of connect.
# Transport established.
def ConnectToDirectory_PENDING_Connected(self, message):
	self.cancel(ar.T1)
	self.connected = message
	self.remote = self.return_address

	# Remote object is ready.
	self.send(UseAddress(self.remote), self.parent_address)
	return CONNECTED

def ConnectToDirectory_PENDING_NotConnected(self, message):
	self.send(message, self.parent_address)
	self.cancel(ar.T1)
	# Attempt failed.
	# No session and no change of status for owner.
	# Schedule another or perhaps end of attempts.
	if self.reschedule():
		return GLARING

	x = ar.Exhausted(message, attempts=self.attempts, started=self.started)
	self.complete(x)

def ConnectToDirectory_PENDING_T1(self, message):
	if self.reschedule():
		return GLARING

	x = ar.Exhausted(message, attempts=self.attempts, started=self.started)
	self.complete(x)

def ConnectToDirectory_PENDING_Stop(self, message):
	# Local termination.
	# Connected could be orphaned here.
	self.complete(ar.Aborted())

# PRODUCING (ProductAccess)
# Waiting for results of connect.
# Transport established.
def ConnectToDirectory_PRODUCING_Connected(self, message):
	self.cancel(ar.T1)
	self.connected = message

	c = self.connect_above
	lookup = ProductLookup(product_name=c.product_name, product_instance=c.product_instance)
	self.reply(lookup)
	return GLANCING

def ConnectToDirectory_PRODUCING_NotConnected(self, message):
	self.send(message, self.parent_address)
	self.cancel(ar.T1)
	# Attempt failed.
	# No session and no change of status for owner.
	# Schedule another or perhaps end of attempts.
	if self.reschedule():
		return GLARING

	x = ar.Exhausted(message, attempts=self.attempts, started=self.started)
	self.complete(x)

def ConnectToDirectory_PRODUCING_T1(self, message):
	if self.reschedule():
		return GLARING

	x = ar.Exhausted(message, attempts=self.attempts, started=self.started)
	self.complete(x)

def ConnectToDirectory_PRODUCING_Stop(self, message):
	# Local termination.
	# Connected could be orphaned here.
	self.complete(ar.Aborted())

def ConnectToDirectory_GLANCING_YourProduct(self, message):
	self.remote = message.address
	# Remote object is ready.
	self.send(UseAddress(self.remote), self.parent_address)
	return CONNECTED

def ConnectToDirectory_GLANCING_Stop(self, message):
	self.send(ar.Close(ar.Aborted()), self.connected.remote_address)
	return CLOSING

# CLOUDY (WideAreaAccess)
# Waiting for results of connect.
# Transport established.
def ConnectToDirectory_CLOUDY_Connected(self, message):
	self.cancel(ar.T1)
	self.connected = message

	c = self.connect_above
	lookup = WideAreaLookup(account_id=c.account_id, directory_id=c.directory_id,
		access_token=c.access_token,
		product_name=c.product_name, product_instance=c.product_instance)
	self.reply(lookup)
	self.start(ar.T2, seconds=4.0)
	return LOOKING

def ConnectToDirectory_CLOUDY_NotConnected(self, message):
	self.send(message, self.parent_address)
	self.cancel(ar.T1)
	# Attempt failed.
	# No session and no change of status for owner.
	# Schedule another or perhaps end of attempts.
	if self.reschedule():
		return GLARING

	x = ar.Exhausted(message, attempts=self.attempts, started=self.started)
	self.complete(x)

def ConnectToDirectory_CLOUDY_T1(self, message):
	# Attempt failed.
	# No session and no change of status for owner.
	# Schedule another or perhaps end of attempts.
	if self.reschedule():
		return GLARING

	x = ar.Exhausted(message, attempts=self.attempts, started=self.started)
	self.complete(x)

def ConnectToDirectory_CLOUDY_Stop(self, message):
	# Local termination.
	# Connected could be orphaned here.
	self.complete(ar.Aborted())

def ConnectToDirectory_LOOKING_WideAreaRedirect(self, message):
	self.redirect = message
	ipp = message.redirect_ipp
	self.trace(f'Redirecting from FOH to WAN ({ipp})')
	encrypted = message.encrypted
	connect(self, ipp, encrypted=encrypted)
		
	self.reply(Close())
	return REDIRECTING

def ConnectToDirectory_LOOKING_Faulted(self, message):
	self.trace(f'Access denied - {message}')
	self.send(Close(message), self.connected.remote_address)
	if self.reschedule():
		return GLARING

	x = ar.Exhausted(message, attempts=self.attempts, started=self.started)
	self.complete(x)

def ConnectToDirectory_LOOKING_T2(self, message):
	value = ar.TimedOut(message)
	self.send(Close(value), self.connected.remote_address)
	if self.reschedule():
		return GLARING

	x = ar.Exhausted(message, attempts=self.attempts, started=self.started)
	self.complete(x)

def ConnectToDirectory_LOOKING_Stop(self, message):
	self.send(Close(ar.Aborted()), self.connected.remote_address)
	return CLOSING

def ConnectToDirectory_REDIRECTING_Connected(self, message):
	self.connected = message
	r = self.redirect
	assignment = WideAreaAssignment(r.directory_id, r.assignment_token)
	self.reply(assignment)
	return ASSIGNING

def ConnectToDirectory_REDIRECTING_NotConnected(self, message):
	self.send(message, self.parent_address)
	if self.reschedule():
		return GLARING

	x = ar.Exhausted(message, attempts=self.attempts, started=self.started)
	self.complete(x)

def ConnectToDirectory_REDIRECTING_T2(self, message):
	# Dropped the original connection and redirect not
	# successful.
	if self.reschedule():
		return GLARING

	x = ar.Exhausted(message, attempts=self.attempts, started=self.started)
	self.complete(x)

def ConnectToDirectory_REDIRECTING_Stop(self, message):
	self.complete(ar.Aborted())

def ConnectToDirectory_ASSIGNING_YourWideArea(self, message):
	self.cancel(ar.T2)
	self.remote = message.address
	# Remote object is ready.
	self.send(UseAddress(self.remote), self.parent_address)
	return CONNECTED

def ConnectToDirectory_ASSIGNING_Closed(self, message):
	# Housekeeping - dropped previous connection.
	return ASSIGNING

def ConnectToDirectory_ASSIGNING_Unknown(self, message):
	if self.reschedule():
		return GLARING

	x = ar.Exhausted(message, attempts=self.attempts, started=self.started)
	self.complete(x)

def ConnectToDirectory_ASSIGNING_T2(self, message):
	value = ar.TimedOut(message)
	self.send(Close(value), self.connected.remote_address)
	if self.reschedule():
		return GLARING

	x = ar.Exhausted(message, attempts=self.attempts, started=self.started)
	self.complete(x)

def ConnectToDirectory_ASSIGNING_Stop(self, message):
	self.complete(ar.Aborted())

# GLARING
# After a failed attempt or after abandoned.
def ConnectToDirectory_GLARING_Unknown(self, message):
	# Non-control message sneaking through.
	self.forward(message, self.group_address, self.return_address)
	return GLARING

def ConnectToDirectory_GLARING_Anything(self, message):
	self.connect_above = message.thing
	self.attempts = 1
	return self.connect()

def ConnectToDirectory_GLARING_GlareTimer(self, message):
	self.attempts += 1
	return self.connect()

def ConnectToDirectory_GLARING_Closed(self, message):
	return GLARING

def ConnectToDirectory_GLARING_Stop(self, message):
	# Drop GlareTimer
	self.complete(ar.Aborted())

# CONNECTED
# Caretaker role. Pass app messages on to owner
# and wait for further control messages.
def ConnectToDirectory_CONNECTED_Unknown(self, message):
	# Normal operation.	Forward app message on to proper target.
	self.forward(message, self.group_address, self.return_address)
	return CONNECTED

def ConnectToDirectory_CONNECTED_Anything(self, message):
	self.connect_above = message.thing
	self.send(Close(), self.remote)
	return RECONNECTING

def ConnectToDirectory_CONNECTED_Abandoned(self, message):
	# Start the retries up again.
	self.started = ar.world_now()
	self.attempts = 0
	self.retry = None
	if self.reschedule():
		# Update the owner that the current session
		# is over.
		self.send(NoAddress(), self.parent_address)
		return GLARING

	x = ar.Exhausted(message, attempts=self.attempts, started=self.started)
	self.complete(x)

def ConnectToDirectory_CONNECTED_Stop(self, message):
	# This object ended by app. Take that as
	# signal to end this session and not activate retries.

	e = ar.Stop() if self.session else Close(ar.Aborted())

	self.send(e, self.remote)
	return CLOSING

def ConnectToDirectory_CONNECTED_Closed(self, message):
	# Local end has sent close to the proxy. Treat this
	# as a short-circuit version of above.
	self.complete(message.value)

#
#
def ConnectToDirectory_RECONNECTING_Abandoned(self, message):
	# Start the retries up again.
	self.started = ar.world_now()
	self.attempts = 0
	self.retry = None
	if self.reschedule():
		self.send(NoAddress(), self.parent_address)
		return GLARING

	x = ar.Exhausted(message, attempts=self.attempts, started=self.started)
	self.complete(x)

def ConnectToDirectory_RECONNECTING_Closed(self, message):
	return ConnectToDirectory_RECONNECTING_Abandoned(self, message)

def ConnectToDirectory_RECONNECTING_Stop(self, message):
	return CLOSING

# CLOSING
def ConnectToDirectory_CLOSING_Unknown(self, message):
	# Normal operation.	Forward app message on to proper target.
	self.forward(message, self.group_address, self.return_address)
	return CLOSING

def ConnectToDirectory_CLOSING_Abandoned(self, message):
	# Terminated by remote before close could get through.
	self.complete(message)

def ConnectToDirectory_CLOSING_Closed(self, message):
	# Completion of CONNECTED-Stop.
	self.complete(message.value)

CONNECT_TO_DIRECTORY_DISPATCH = {
	INITIAL: (
		(ar.Start,), ()
	),
	DISABLED: (
		(ar.Anything, ar.Stop), ()
	),
	PENDING: (
		(Connected, NotConnected, ar.T1, ar.Stop), (ar.Anything,)
	),
	PRODUCING: (
		(Connected, NotConnected, ar.T1, ar.Stop), (ar.Anything,)
	),
	GLANCING: (
		(YourProduct, ar.Stop), (ar.Anything,)
	),
	CLOUDY: (
		(Connected, NotConnected, ar.T1, ar.Stop), (ar.Anything,)
	),
	LOOKING: (
		(WideAreaRedirect, ar.Faulted, ar.T2, ar.Stop), (ar.Anything,)
	),
	REDIRECTING: (
		(Connected, NotConnected, ar.T2, ar.Stop), (ar.Anything,)
	),
	ASSIGNING: (
		(YourWideArea, Closed, ar.Unknown, ar.T2, ar.Stop), (ar.Anything,)
	),
	GLARING: (
		(ar.Unknown, ar.Anything, Closed, GlareTimer, ar.Stop), ()
	),
	CONNECTED: (
		(ar.Unknown, ar.Anything, Abandoned, Closed, ar.Stop), ()
	),
	RECONNECTING: (
		(Closed, Abandoned, ar.Stop), ()
	),
	CLOSING: (
		(ar.Unknown, Abandoned, Closed), ()
	),
}

ar.bind(ConnectToDirectory, CONNECT_TO_DIRECTORY_DISPATCH)
